'use strict';

var fs = require('fs');
var typescriptLogging = require('typescript-logging');
var os = require('os');
var p = require('path');

function _interopNamespaceDefault(e) {
  var n = Object.create(null);
  if (e) {
    Object.keys(e).forEach(function (k) {
      if (k !== 'default') {
        var d = Object.getOwnPropertyDescriptor(e, k);
        Object.defineProperty(n, k, d.get ? d : {
          enumerable: true,
          get: function () { return e[k]; }
        });
      }
    });
  }
  n.default = e;
  return Object.freeze(n);
}

var fs__namespace = /*#__PURE__*/_interopNamespaceDefault(fs);
var os__namespace = /*#__PURE__*/_interopNamespaceDefault(os);
var p__namespace = /*#__PURE__*/_interopNamespaceDefault(p);

/**
 * Convert given fileSize to bytes
 * @param fileSize File size to convert
 */
function fileSizeToBytes(fileSize) {
    switch (fileSize.unit) {
        case "MegaBytes":
            return 1024 * 1024 * fileSize.value;
        case "KiloBytes":
            return 1024 * fileSize.value;
        default:
            throw new Error(`Type '${fileSize.unit}' is not supported.`);
    }
}

/**
 * Queue which has maximum size of elements. When the maximum is reached when adding a new element, it will
 * remove the old element. This is to assure we don't buffer an enormous amount of messages (causing memory problems).
 */
class Queue {
    constructor(maxElementsAllowed) {
        this._queue = [];
        this._maxElementsAllowed = maxElementsAllowed;
    }
    /**
     * Add a new element, when the queue is full, the oldest element is *removed* before this element is added.
     * @param {T} value
     */
    add(value) {
        if (this._queue.length === this._maxElementsAllowed) {
            this._queue.splice(0, 1);
        }
        this._queue.push(value);
    }
    peekHead() {
        if (this._queue.length === 0) {
            return null;
        }
        return this._queue[0];
    }
    popHead() {
        if (this._queue.length === 0) {
            return null;
        }
        return this._queue.shift();
    }
    get size() {
        return this._queue.length;
    }
    /** Drain queue and returns all data */
    drain() {
        const result = [...this._queue];
        this._queue.length = 0;
        return result;
    }
    /** Get a copy of the current data */
    getData() {
        return [...this._queue];
    }
}

/**
 * The writer writes to a file. It has a buffer (queue) to throttle writing to the file (backpressure). The queue will throw away old messages
 * if IO can't keep up, this would only occur if there is an insane amount of logging.
 */
class NodeLogWriter {
    constructor(path, encoding, maxFileSize, startFileSize, maxQueuedMessages, initialQueueBuffer, onLimitReached, streamCallBack) {
        this._limitReached = false;
        this._writerIsProcessing = false;
        this._path = path;
        this._encoding = encoding;
        this._streamCallBack = streamCallBack;
        this._maxFileSizeInBytes = fileSizeToBytes(maxFileSize);
        this._bytesWritten = startFileSize;
        this._onLimitReached = onLimitReached;
        this.write = this.write.bind(this);
        this.close = this.close.bind(this);
        this.getRemainingQueueBuffer = this.getRemainingQueueBuffer.bind(this);
        this.drainRemainingBuffer = this.drainRemainingBuffer.bind(this);
        this.onError = this.onError.bind(this);
        this.onFinished = this.onFinished.bind(this);
        this.onClose = this.onClose.bind(this);
        this.write = this.write.bind(this);
        this.processQueue = this.processQueue.bind(this);
        this._writeQueue = new Queue(maxQueuedMessages);
        for (const value of initialQueueBuffer) {
            this._writeQueue.add(value);
        }
        if (initialQueueBuffer.length > 0) {
            /** Start processing as there is data already. */
            setTimeout(() => this.processQueue(), 10);
        }
        this._log = typescriptLogging.$internal.getInternalLogger("node.NodeLogWriter");
        this._writeStream = this.getStream();
    }
    /**
     *  Add data to write to the queue.
     */
    write(message) {
        const sizeString = Buffer.byteLength(message, this._encoding);
        // Sanity check
        if (sizeString > this._maxFileSizeInBytes) {
            throw new Error(`Attempting to write a string which is larger than the maximum file size of a file ${this._maxFileSizeInBytes}, please check your code and/or configuration`);
        }
        this._writeQueue.add(message);
        this.processQueue();
    }
    drainRemainingBuffer() {
        return this._writeQueue.drain();
    }
    getRemainingQueueBuffer() {
        return this._writeQueue.getData();
    }
    isQueueEmpty() {
        return this._writeQueue.size === 0;
    }
    close() {
        if (!this._writeStream) {
            return;
        }
        this._log.debug(() => `Forcefully closing: ${this._path}`);
        this._writeStream.end();
        this._writeStream = undefined;
        this._limitReached = true;
    }
    processQueue() {
        if (this._limitReached) {
            return;
        }
        if (!this._writeStream) {
            throw new Error("Cannot write, stream is not available.");
        }
        /* Try to write max 20 messages as long as the writer accepts it. */
        for (let count = 0; count < 20; count++) {
            if (!this._writerIsProcessing && this._writeQueue.size > 0) {
                const nextString = this._writeQueue.peekHead();
                const sizeString = Buffer.byteLength(nextString, this._encoding);
                if (this._bytesWritten + sizeString > this._maxFileSizeInBytes) {
                    this._limitReached = true;
                    this._writerIsProcessing = false;
                    this._onLimitReached();
                    return;
                }
                this._writeQueue.popHead();
                this._bytesWritten += sizeString;
                const success = this._writeStream.write(nextString, () => {
                    this._writerIsProcessing = false;
                    this.processQueue();
                });
                if (!success) {
                    this._writerIsProcessing = true;
                    // If write returns false, wait for the 'drain' event before writing more data
                    this._writeStream.once("drain", () => {
                        this._writerIsProcessing = false;
                        this.processQueue();
                    });
                    // Break, we cannot write more.
                    break;
                }
            }
        }
    }
    getStream() {
        this._log.debug(() => `Opening stream for path '${this._path}' using encoding '${this._encoding}'`);
        this._writeStream = fs.createWriteStream(this._path, { encoding: this._encoding, flags: "a" });
        this._writeStream.once("error", this.onError);
        this._writeStream.once("finish", this.onFinished);
        this._writeStream.once("close", this.onClose);
        return this._writeStream;
    }
    onError(error) {
        this._log.error(() => `Unexpected error for: ${this._path}`, error);
        if (this._streamCallBack && this._streamCallBack.onError) {
            this._streamCallBack.onError(error);
        }
        /* Don't try to recover just destroy ourselves */
        if (this._writeStream) {
            this._writeStream.destroy();
            this._writeStream = undefined;
        }
    }
    onFinished() {
        this._log.debug(() => `onFinished for: ${this._path}`);
        if (this._streamCallBack && this._streamCallBack.onFinished) {
            this._streamCallBack.onFinished(this._path);
        }
    }
    onClose() {
        this._log.debug(() => `onClose for: ${this._path}`);
        if (this._streamCallBack && this._streamCallBack.onClose) {
            this._streamCallBack.onClose(this._path);
        }
    }
}

const minQueueMessages = 100;
const maxQueuedMessages = 1000000;
const defaultQueuedMessages = 5000;
/**
 * Base class for node channels which contains all the logic around retention and where to write to.
 */
class AbstractNodeChannel {
    constructor(retentionStrategy, options) {
        this._initialized = false;
        this._writerClosed = true;
        this._retentionStrategy = retentionStrategy;
        this._maxQueuedMessages = options && options.maxQueuedMessages ? options.maxQueuedMessages : defaultQueuedMessages;
        if (this._maxQueuedMessages < minQueueMessages || this._maxQueuedMessages > maxQueuedMessages) {
            throw new Error(`Maximum of queued messages must be between ${minQueueMessages} and ${maxQueuedMessages}.`);
        }
        this._onRollOver = options && options.onRollOver ? options.onRollOver : undefined;
        this.closeWriter = this.closeWriter.bind(this);
        this.isWriterQueueEmpty = this.isWriterQueueEmpty.bind(this);
        this.writeMessage = this.writeMessage.bind(this);
        this.initialize = this.initialize.bind(this);
        this.nextWriter = this.nextWriter.bind(this);
        this.onLimitReached = this.onLimitReached.bind(this);
        this.onStreamError = this.onStreamError.bind(this);
        this.onStreamFinished = this.onStreamFinished.bind(this);
        this.onStreamClose = this.onStreamClose.bind(this);
        this.onRetentionRollOver = this.onRetentionRollOver.bind(this);
    }
    /**
     * Forcefully close the underlying writer. Should not be used by end users, for testing purposes only. May roll over if needed.
     */
    closeWriter() {
        if (this._writer) {
            this._writer.close();
        }
    }
    /**
     * Check whether the current writers queue is empty. Should not be used by end users, for testing purposes only.
     * @returns {boolean}
     */
    isWriterQueueEmpty() {
        if (!this._writer) {
            throw new Error("No writer present");
        }
        return this._writer.isQueueEmpty();
    }
    /**
     * For testing purposes only, provides access to some stream events.
     * Will be applied next time a new stream is opened (not to current stream if any is open).
     *
     * @param callbacks The custom callback, when undefined unsets the callback.
     */
    setStreamCallBacks(callbacks) {
        this._streamCallBacks = callbacks;
    }
    /**
     * Subclasses must NOT override this function, instead implement the write(...) function of the channel you implement (RawLogChannel / LogChannel), create the message
     * there and then call this method to write the actual data.
     * @param msg Message to write
     * @protected
     */
    writeMessage(msg) {
        var _a;
        if (!this._initialized) {
            this.initialize();
        }
        if (this._writerClosed) {
            this.nextWriter();
        }
        (_a = this._writer) === null || _a === void 0 ? void 0 : _a.write(msg);
    }
    initialize() {
        this._retentionStrategy.initialize(this.onRetentionRollOver);
        this._initialized = true;
    }
    nextWriter() {
        var _a;
        /* We roll over if there is a writer, as it means this class decided it needs a rollover, otherwise it's startup let the strategy decide instead. */
        const mustRollOver = this._writer !== undefined;
        const initialBuffer = this._writer !== undefined ? this._writer.drainRemainingBuffer() : [];
        (_a = this._writer) === null || _a === void 0 ? void 0 : _a.close();
        const [nextFile, currentFileSize] = this._retentionStrategy.nextFile(mustRollOver);
        this._writer = new NodeLogWriter(nextFile, this._retentionStrategy.encoding, this._retentionStrategy.maxFileSize, currentFileSize, this._maxQueuedMessages, initialBuffer, this.onLimitReached, {
            onFinished: this.onStreamFinished,
            onError: this.onStreamError,
            onClose: this.onStreamClose
        });
        this._writerClosed = false;
        return this._writer;
    }
    onLimitReached() {
        var _a;
        (_a = this._writer) === null || _a === void 0 ? void 0 : _a.close();
    }
    onStreamFinished(path) {
        if (this._streamCallBacks && this._streamCallBacks.onFinished) {
            this._streamCallBacks.onFinished(path);
        }
    }
    onStreamClose(path) {
        this._writerClosed = true;
        if (this._streamCallBacks && this._streamCallBacks.onClose) {
            this._streamCallBacks.onClose(path);
        }
        /* Data dit not fit in last writer and is available, so go to work again. */
        if (!this.isWriterQueueEmpty()) {
            this.nextWriter();
        }
    }
    onStreamError(error) {
        /*
          This also closes the current errored stream, we will try the next instead.
         */
        try {
            this.nextWriter();
        }
        finally {
            if (this._streamCallBacks && this._streamCallBacks.onError) {
                this._streamCallBacks.onError(error);
            }
        }
    }
    onRetentionRollOver(path) {
        /* We only care if there is a rollover specified by the end user. */
        if (this._onRollOver) {
            /* Spawn it a little later, so we don't block our current event. */
            setTimeout(() => {
                if (this._onRollOver) {
                    this._onRollOver(path);
                }
            }, 10);
        }
    }
}

/**
 * Node Channel implementation of type LogChannel.
 */
class NodeLogChannel extends AbstractNodeChannel {
    constructor(retentionStrategy, options) {
        super(retentionStrategy, options);
    }
    get type() {
        return "LogChannel";
    }
    write(msg) {
        const fullMsg = msg.message + (msg.error ? `\n${msg.error}` : "") + os__namespace.EOL;
        this.writeMessage(fullMsg);
    }
}

/**
 * Returns Dirent (files) within given directory that match given matcher function
 * @param directory Directory to check (must exist)
 * @param encoding Encoding to use
 * @param fnMatch Called for each file/directory
 */
function listFiles(directory, encoding, fnMatch) {
    try {
        return fs__namespace.readdirSync(directory, { encoding, withFileTypes: true })
            .filter(fnMatch);
    }
    catch (e) {
        if (e instanceof Error) {
            throw e;
        }
        throw new Error(`Failed to list files for directory '${directory}'.`);
    }
}

/**
 * Implements the max files retention strategy.
 */
class RetentionStrategyMaxFiles {
    constructor(options) {
        this._allowedFileNamesShort = new Set();
        this._directory = options.directory;
        this._encoding = options.encoding;
        this._namePrefix = options.namePrefix;
        this._extension = options.extension;
        this._maxFileSize = options.maxFileSize;
        this._maxFileSizeBytes = fileSizeToBytes(options.maxFileSize);
        this._maxFiles = options.maxFiles;
        this._log = typescriptLogging.$internal.getInternalLogger("node.RetentionMaxFiles");
        const exists = fs__namespace.existsSync(this._directory);
        if (!exists) {
            const created = fs__namespace.mkdirSync(this._directory, { recursive: true });
            if (!created) {
                throw new Error(`Failed to create directory '${this._directory}', is the path valid?`);
            }
        }
        try {
            const fileInfo = fs__namespace.statSync(this._directory);
            if (!fileInfo.isDirectory()) {
                // noinspection ExceptionCaughtLocallyJS
                throw new Error(`Directory '${this._directory}' does exist but is not a directory, or cannot be accessed.`);
            }
        }
        catch (e) {
            if (e instanceof Error) {
                throw e;
            }
            throw new Error(`Failed to check that directory '${this._directory}' is indeed a directory`);
        }
    }
    get encoding() {
        return this._encoding;
    }
    get maxFileSize() {
        return this._maxFileSize;
    }
    initialize(onRollOver) {
        this._onRollOver = onRollOver;
        for (let i = 1; i <= this._maxFiles; i++) {
            this._allowedFileNamesShort.add(this.getShortFileName(i));
        }
        const currentFile = this.determineCurrentFile();
        if (currentFile) {
            this._lastFileShortName = currentFile.name;
        }
    }
    nextFile(mustRollOver) {
        const logFilesFound = this.getLogFiles();
        if (logFilesFound.length === 0) {
            const result = this.getLogFileName(1);
            this._lastFileShortName = this.getShortFileName(1);
            return [result, 0];
        }
        const lastFile = logFilesFound.find(f => f.name === this._lastFileShortName);
        if (!lastFile) {
            /* Seems the file was removed in between, try again .. */
            return this.nextFile(mustRollOver);
        }
        /* The check on size is only there for startup, while running it will never be up-to-date as files won't be flushed by node yet (due to events still need to be processed by node) */
        if (mustRollOver || lastFile.sizeInBytes >= this._maxFileSizeBytes) {
            /* We are forced to rollover, or the file is full, so go to the next file (1 up) */
            const result = this.rolloverFile(lastFile);
            this._lastFileShortName = result[1];
            return [result[0], result[2]];
        }
        /* Existing file is good to go */
        this._lastFileShortName = lastFile.name;
        return [lastFile.filePath, lastFile.sizeInBytes];
    }
    rolloverFile(lastFile) {
        var _a;
        let numValue = parseInt(lastFile.name.substring(this._namePrefix.length, lastFile.name.length - this._extension.length), 10);
        if (numValue === this._maxFiles) {
            numValue = 1;
        }
        else {
            numValue = numValue + 1;
        }
        const newFileName = this.getLogFileName(numValue);
        try {
            if (fs__namespace.existsSync(newFileName)) {
                fs__namespace.unlinkSync(newFileName);
            }
        }
        catch (e) {
            this._log.warn(() => `Failed to delete existing log file ${newFileName}, rollover may fail.`, e);
        }
        finally {
            (_a = this._onRollOver) === null || _a === void 0 ? void 0 : _a.call(this, lastFile.filePath);
        }
        return [newFileName, this.getShortFileName(numValue), 0];
    }
    /**
     * Get all matching log files
     * @private
     */
    getLogFiles() {
        return listFiles(this._directory, this._encoding, f => f.isFile() && this._allowedFileNamesShort.has(f.name))
            .map(file => {
            const filePath = this._directory + p__namespace.sep + file.name;
            try {
                const fileInfo = fs__namespace.statSync(filePath);
                return { lastModified: fileInfo.mtimeMs, sizeInBytes: fileInfo.size, filePath, name: file.name, valid: true };
            }
            catch (e) {
                this._log.warn(() => `Failed to get file stats for ${filePath}`, e);
                return { lastModified: 0, sizeInBytes: 0, filePath: "", name: "", valid: false };
            }
        })
            .filter(v => v.valid);
    }
    getLogFileName(value) {
        return this._directory + p__namespace.sep + this._namePrefix + value + this._extension;
    }
    getShortFileName(value) {
        return this._namePrefix + value + this._extension;
    }
    determineCurrentFile() {
        const logFilesFound = this.getLogFiles();
        if (logFilesFound.length === 0) {
            return undefined;
        }
        /**
         * Find the last file in use.
         */
        return logFilesFound.reduce((lhs, rhs) => {
            const diff = lhs.lastModified - rhs.lastModified;
            if (diff > 0) {
                return lhs;
            }
            if (diff < 0) {
                return rhs;
            }
            return lhs;
        });
    }
}

/**
 * Node Channel implementation of type RawLogChannel.
 */
class NodeRawLogChannel extends AbstractNodeChannel {
    constructor(retentionStrategy, writeRawLogMessage, options) {
        super(retentionStrategy, options);
        this._writeRawLogMessage = writeRawLogMessage;
    }
    get type() {
        return "RawLogChannel";
    }
    write(msg, formatArg) {
        const result = this._writeRawLogMessage(msg, formatArg);
        this.writeMessage(result);
    }
}

/**
 * Allows creation of various node channels that can be used to log for node to files.
 */
class NodeChannelProvider {
    /**
     * Create a new LogChannel which will use given RetentionStrategy, you can use NodeChannelProvider.createRetentionStrategyMaxFiles(..) to create
     * the RetentionStrategy (or implement a custom one).
     * @param retentionStrategy The retention type
     * @param options Additional channel options
     */
    static createLogChannel(retentionStrategy, options) {
        return new NodeLogChannel(retentionStrategy, options);
    }
    /**
     * Create a new RawLogChannel  which will use given RetentionStrategy, you can use NodeChannelProvider.createRetentionStrategyMaxFiles(..) to create
     * the RetentionStrategy (or implement a custom one).
     * @param retentionStrategy The retention type
     * @param writeRawLogMessage Function that does format the raw message and returns it as a string, this string is then logged in the file.
     * @param options Additional channel options
     */
    static createRawLogChannel(retentionStrategy, writeRawLogMessage, options) {
        return new NodeRawLogChannel(retentionStrategy, writeRawLogMessage, options);
    }
    /**
     * Create a new RetentionStrategy based on given options. For details please {@see RetentionStrategyMaxFilesOptions}
     * @param options
     */
    static createRetentionStrategyMaxFiles(options) {
        return new RetentionStrategyMaxFiles({
            directory: options.directory,
            maxFiles: options.maxFiles ? options.maxFiles : 10,
            maxFileSize: options.maxFileSize ? options.maxFileSize : { value: 10, unit: "MegaBytes" },
            encoding: options.encoding ? options.encoding : "utf-8",
            extension: options.extension ? options.extension : ".log",
            namePrefix: options.namePrefix ? options.namePrefix : "application",
        });
    }
}

exports.NodeChannelProvider = NodeChannelProvider;
exports.fileSizeToBytes = fileSizeToBytes;
//# sourceMappingURL=typescript-logging-node.js.map
